from __future__ import annotations

import pytest

from ..itemize import itemize

text_chunks = ['abc', 'def\n012', '\n', '\n']
byte_chunks = [chunk.encode() for chunk in text_chunks]
text_chunks_other = [chunk.replace('\n', '\r\n') for chunk in text_chunks]
byte_chunks_other = [chunk.encode() for chunk in text_chunks_other]


@pytest.mark.parametrize(
    ('input_chunks', 'separator'),
    [
        (text_chunks, '\n'),
        (byte_chunks, b'\n'),
        (text_chunks_other, '\r\n'),
        (byte_chunks_other, b'\r\n'),
    ],
)
def test_assembling_and_splitting(input_chunks, separator):
    empty = input_chunks[0][:0]

    expected_items = 3

    r = tuple(itemize(input_chunks, None, keep_ends=True))
    assert len(r) == expected_items
    assert empty.join(r) == empty.join(input_chunks)

    r = tuple(itemize(input_chunks, sep=separator, keep_ends=True))
    assert len(r) == expected_items
    assert empty.join(r) == empty.join(input_chunks)

    r = tuple(itemize(input_chunks, sep=separator))
    assert len(r) == expected_items
    assert empty.join(r) == empty.join(input_chunks).replace(separator, empty)

    r = tuple(itemize(input_chunks + input_chunks[:1], sep=separator, keep_ends=True))
    assert len(r) == expected_items + 1
    assert r[3] == input_chunks[0]



"""Exception raised on a failed command execution"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import os

import signal
import sys


class CommandError(RuntimeError):
    """Raised when a subprocess execution fails (non-zero exit)

    Key class attributes are aligned with the ``CalledProcessError`` exception
    of the ``subprocess`` module of the Python standard library. However,
    this class is not a subclass of ``CalledProcessError``, but a subclass
    of ``RuntimeError``.

    At minimum, the command with the failed execution must be provided
    as an argument. A number of additional information items can be provided
    in addition to enable more comprehensive reporting on execution failures.

    As an addition to the ``CalledProcessError`` arguments, a ``msg`` parameter
    is supported which can be used to include contextual information on the
    command execution, for example why a command execution was attempted, or
    under which particular circumstances. A possible implementation pattern
    is::

        try:
            # some call raises an original exception
            raise CommandError('mycmd')
        except CommandError as e:
            # a wrapping try/except can be used to add context info
            # or a hint on a probably cause to the exception
            e.msg = 'context info or hint'
            raise e
    """

    def __init__(
        self,
        cmd: str | list[str],
        msg: str = '',
        returncode: int | None = None,
        stdout: str | bytes = '',
        stderr: str | bytes = '',
        cwd: str | os.PathLike | None = None,
    ) -> None:
        RuntimeError.__init__(self, msg)
        self.cmd = cmd
        self.msg = msg
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr
        self.cwd = cwd

    def __str__(self) -> str:
        # we report the command verbatim, in exactly the form that it has
        # been given to the exception. Previously implementation have
        # beautified output by joining list-format commands with shell
        # quoting. However that implementation assumed that the command
        # actually run locally. In practice, CommandError is also used
        # to report on remote command execution failure. Reimagining
        # quoting and shell conventions based on assumptions is confusing.
        to_str = f'Command {self.cmd!r}'
        if self.returncode and self.returncode < 0:
            try:
                to_str += f' died with {signal.Signals(-self.returncode).name}'
            except ValueError:
                to_str += f' died with unknown signal {-self.returncode}'
        elif self.returncode:
            to_str += f' returned non-zero exit status {self.returncode}'
        else:
            to_str += ' errored with unknown exit status'
        if self.cwd:
            # only if not under standard PWD
            to_str += f' at CWD {self.cwd}'
        if self.msg:
            # typically a command error has no specific idea
            # but we support it, because CommandError derives
            # from RuntimeError which has this feature.
            to_str += f' [{self.msg}]'

        if not self.stderr:
            return to_str

        # make an effort to communicate stderr
        stderr = ''
        if isinstance(self.stderr, bytes):
            # assume that the command output matches the local system
            # encoding
            try:
                # we need to try conversion on the full bytestring to
                # avoid alignment issues with random splits
                stderr = self.stderr.decode(sys.getdefaultencoding())
            except UnicodeDecodeError:
                # we tried, we failed, sorry
                # we are not guessing other encodings. If it doesn't
                # match the system encoding, it is somewhat unlikely
                # to be an informative error message.
                stderr = f'<undecodable {truncate_bytes(self.stderr)}>'
        else:
            stderr = self.stderr

        to_str += f' [stderr: {truncate_str(stderr, (60, 0))}]'

        return to_str

    def __repr__(self) -> str:
        descr = f'{self.__class__.__name__}({self.cmd!r}'
        for kwarg, (val, default) in {
            'msg': (self.msg, ''),
            'returncode': (self.returncode, None),
            'stdout': (self.stdout, ''),
            'stderr': (self.stderr, ''),
            'cwd': (self.cwd, None),
        }.items():
            if val == default:
                continue
            if kwarg in ('stdout', 'stderr'):
                if TYPE_CHECKING:
                    assert isinstance(val, (str, bytes))
                if isinstance(val, bytes):
                    descr += f", {kwarg}=b'<{truncate_bytes(val)}>'"
                else:
                    descr += f', {kwarg}={truncate_str(val)!r}'
            else:
                descr += f', {kwarg}={val!r}'
        descr += ')'
        return descr


def truncate_bytes(data: bytes) -> str:
    return f'{len(data)} bytes'


def truncate_str(text: str, keep: tuple[int, int] = (20, 20)) -> str:
    # truncation like done below only actually shortens beyond
    # 60 chars input length
    front, back = keep
    if len(text) < (front + back + 14):
        # stringify only
        return f'{text}'
    return (
        f"{text[:front]}<... +{len(text) - front - back} chars>"
        f"{text[-back:] if back > 0 else ''}"
    )



"""High-level utilties for execution of subprocesses

This module provides relevant components for subprocess execution.
Execution errors are communicated with the
:class:`~datasalad.runners.CommandError` exception.

.. currentmodule:: datasalad.runners
.. autosummary::
   :toctree: generated

   CommandError
   iter_subproc
"""

__all__ = ['CommandError', 'iter_subproc']

from .exception import CommandError
from .iter_subproc import iter_subproc



from __future__ import annotations

import pytest

from ..align_pattern import align_pattern


@pytest.mark.parametrize(
    ('data_chunks', 'pattern', 'expected'),
    [
        (['a', 'b', 'c', 'd', 'e'], 'abc', ['abc', 'de']),
        (['a', 'b', 'c', 'a', 'b', 'c'], 'abc', ['abc', 'abc']),
        # Ensure that unaligned pattern prefixes are not keeping data chunks short.
        (['a', 'b', 'c', 'dddbbb', 'a', 'b', 'x'], 'abc', ['abc', 'dddbbb', 'abx']),
        # Expect that a trailing minimum length-chunk that ends with a pattern
        # prefix is not returned as data, but as remainder, if it is not the final
        # chunk.
        (['a', 'b', 'c', 'd', 'a'], 'abc', ['abc', 'da']),
        # Expect the last chunk to be returned as data, if final is True, although
        # it ends with a pattern prefix. If final is false, the last chunk will be
        # returned as a remainder, because it ends with a pattern prefix.
        (['a', 'b', 'c', 'dddbbb', 'a'], 'abc', ['abc', 'dddbbb', 'a']),
        (['a', 'b', 'c', '9', 'a'], 'abc', ['abc', '9a']),
    ],
)
def test_pattern_processor(data_chunks, pattern, expected):
    assert expected == list(align_pattern(data_chunks, pattern=pattern))


def test_newline_matches():
    pattern = b'----datalad-end-marker-3654137433-rekram-dne-dalatad----\n'
    chunk1 = b'Have a lot of fun...\n----datalad-end-marker-3654137433-r'
    chunk2 = b'e'
    chunk3 = b'kram-dne-dalatad----\n'
    result = list(align_pattern([chunk1, chunk2, chunk3], pattern))
    assert result == [chunk1 + chunk2 + chunk3]



"""Various iterators, e.g., for subprocess pipelining and output processing

.. currentmodule:: datasalad.itertools
.. autosummary::
   :toctree: generated

    align_pattern
    decode_bytes
    itemize
    load_json
    load_json_with_flag
    route_out
    route_in
"""

__all__ = [
    'align_pattern',
    'decode_bytes',
    'itemize',
    'load_json',
    'load_json_with_flag',
    'StoreOnly',
    'route_in',
    'route_out',
]

from .align_pattern import align_pattern
from .decode_bytes import decode_bytes
from .itemize import itemize
from .load_json import (
    load_json,
    load_json_with_flag,
)
from .reroute import (
    StoreOnly,
    route_in,
    route_out,
)



"""Functions that allow to route data around upstream iterator"""

from __future__ import annotations

from typing import (
    Any,
    Callable,
    Generator,
    Iterable,
)

__all__ = ['StoreOnly', 'route_in', 'route_out']


class StoreOnly:
    pass


def route_out(
    iterable: Iterable,
    data_store: list,
    splitter: Callable[[Any], tuple[Any, Any]],
) -> Generator:
    """Route data around the consumer of this iterable

    :func:`route_out` allows its user to:

     1. store data that is received from an iterable,
     2. determine whether this data should be yielded to a consumer of
        ``route_out``, by calling :func:`splitter`.

    To determine which data is to be yielded to the consumer and which data
    should only be stored but not yielded, :func:`route_out` calls
    :func:`splitter`. :func:`splitter` is called for each item of the input
    iterable, with the item as sole argument. The function should return a
    tuple of two elements. The first element is the data that is to be
    yielded to the consumer. The second element is the data that is to be
    stored in the list ``data_store``. If the first element of the tuple is
    ``datasalad.itertools.StoreOnly``, no data is yielded to the
    consumer.

    :func:`route_in` can be used to combine data that was previously
    stored by :func:`route_out` with the data that is yielded by
    :func:`route_out` and with the data the was not processed, i.e. not yielded
    by :func:`route_out`.

    The items yielded by :func:`route_in` will be in the same
    order in which they were passed into :func:`route_out`, including the
    items that were not yielded by :func:`route_out` because :func:`splitter`
    returned ``StoreOnly`` in the first element of the result-tuple.

    The combination of the two functions :func:`route_out` and :func:`route_in`
    can be used to "carry" additional data along with data that is processed by
    iterators. And it can be used to route data around iterators that cannot
    process certain data.

    For example, a user has an iterator to divide the number ``2`` by all
    numbers in a list. The user wants the iterator to process all numbers in a
    divisor list, except from zeros, In this case :func:`route_out` and
    :func:`route_in` can be used as follows:

    .. code-block:: python

        from math import nan
        from datasalad.itertools import route_out, route_in, StoreOnly


        def splitter(divisor):
            # if divisor == 0, return `StoreOnly` in the first element of the
            # result tuple to indicate that route_out should not yield this
            # element to its consumer
            return (StoreOnly, divisor) if divisor == 0 else (divisor, divisor)


        def joiner(processed_data, stored_data):
            #
            return nan if processed_data is StoreOnly else processed_data


        divisors = [0, 1, 0, 2, 0, 3, 0, 4]
        store = list()
        r = route_in(
            (2.0 / x for x in route_out(divisors, store, splitter)), store, joiner
        )
        print(list(r))

    The example about will print ``[nan, 2.0, nan, 1.0, nan, 0.6666666666666666, nan, 0.5]``.

    Parameters
    ----------
    iterable: Iterable
        The iterable that yields the input data
    data_store: list
        The list that is used to store the data that is routed out
    splitter: Callable[[Any], tuple[Any, Any | None]]
        The function that is used to determine which part of the input data,
        if any, is to be yielded to the consumer and which data is to
        be stored in the list ``data_store``.
        The function is called for each item of
        the input iterable with the item as sole argument. It should return a
        tuple of two elements. If the first element is not
        ``datasalad.itertools.StoreOnly``, it is yielded to the consumer.
        If the first element is ``datasalad.itertools.StoreOnly``,
        nothing is yielded to the consumer. The second element is stored in the
        list ``data_store``.
        The cardinality of ``data_store`` will be the same as the cardinality of
        the input iterable.
    """
    for item in iterable:
        data_to_process, data_to_store = splitter(item)
        data_store.append((data_to_process, data_to_store))
        if data_to_process is not StoreOnly:
            yield data_to_process


def route_in(
    iterable: Iterable, data_store: list, joiner: Callable[[Any, Any], Any]
) -> Generator:
    """Yield previously rerouted data to the consumer

    This function is the counter-part to :func:`route_out`. It takes the iterable
    ``iterable`` and a data store given in ``data_store`` and yields items
    in the same order in which :func:`route_out` received them from its
    underlying iterable (using the same data store). This includes items that
    were not yielded by :func:`route_out`, but only stored.

    :func:`route_in` uses :func:`joiner`-function to determine how stored and
    optionally processed data should be joined into a single item, which is
    then yielded by :func:`route_in`.
    :func:`route_in` calls :func:`joiner` with a 2-tuple. The first
    element of the tuple is either ``datasalad.itertools.StoreOnly`` or the
    next item from the underlying iterator. The second element is the data
    that was stored in the data store. The result of :func:`joiner` which will
    be yielded by :func:`route_in`.

    This module provides a standard joiner-function: :func:`join_with_list`
    that works with splitter-functions that return a list as second element of
    the result tuple.

    The cardinality of ``iterable`` must match the number of processed data
    elements in the data store. The output cardinality of :func:`route_in` will
    be the cardinality of the input iterable of the corresponding
    :func:`route_out`-call. Given the following code:

    .. code-block:: python

        store_1 = list()
        route_in(
            some_generator(route_out(input_iterable, store_1, splitter_1)),
            store_1,
            joiner_1,
        )

    :func:`route_in` will yield the same number of elements as ``input_iterable``.
    But, the number of elements processed by ``some_generator`` is determined by
    the :func:`splitter_1` in :func:`route_out`, i.e. by the number of
    :func:`splitter_1`-results that have don't have
    ``datasalad.itertools.don_process`` as first element.

    Parameters
    ----------
    iterable: Iterable
        The iterable that yields the input data.
    data_store: list
        The list from which the data that is to be "routed in" is read.
    joiner: Callable[[Any, Any], Any]
        A function that determines how the items that are yielded by
        ``iterable`` should be combined with the corresponding data from
        ``data_store``, in order to yield the final result.
        The first argument to ``joiner`` is the item that is yielded by
        ``iterable``, or ``datasalad.itertools.StoreOnly`` if no data
        was processed in the corresponding step. The second argument is the
        data that was stored in ``data_store`` in the corresponding step.

    Raises
    ------
    ValueError
      When the "route-in items" from ``data_store`` do not match those
      given by ``iterable``.
    """
    for element in iterable:
        processed, stored = data_store.pop(0)
        # yield stored-only content until we find an item that was processed
        while processed is StoreOnly:
            yield joiner(processed, stored)
            processed, stored = data_store.pop(0)
        yield joiner(element, stored)
    # we reached the end of the incoming iterable.
    # this means that we must not find any remaining items in `data_store`
    # that indicate that they would have a corresponding item in the
    # iterable (processed is not StoreOnly)
    for processed, stored in data_store:
        if processed is not StoreOnly:
            msg = (
                'iterable did not yield matching item for route-in item, '
                'cardinality mismatch?'
            )
            raise ValueError(msg)
        yield joiner(processed, stored)
    # rather than pop() in the last loop, we just yielded from the list
    # now this information is no longer needed
    del data_store[:]



"""Functions that yield JSON objects converted from input items"""

from __future__ import annotations

import json
from typing import (
    Any,
    Generator,
    Iterable,
)

__all__ = ['load_json', 'load_json_with_flag']


def load_json(
    iterable: Iterable[bytes | str],
) -> Generator[Any, None, None]:
    """Convert items yielded by ``iterable`` into JSON objects and yield them

    This function fetches items from the underlying
    iterable. The items are expected to be ``bytes``, ``str``, or ``bytearry``,
    and contain one JSON-encoded object. Items
    are converted into a JSON-object, by feeding them into
    ``json.loads``.

    On successful conversion to a JSON-object, ``load_json`` will yield the
    resulting JSON-object. If the conversion to a JSON-object fails,
    ``load_json`` will raise a ``json.decoder.JSONDecodeError``:

    .. code-block:: python

        >>> from datasalad.itertools import load_json, load_json_with_flag
        >>> tuple(load_json(['{"a": 1}']))
        ({'a': 1},)
        >>> tuple(load_json(['{"c": 3']))   # Faulty JSON-encoding, doctest: +SKIP
        Traceback (most recent call last):
            ...
        json.decoder.JSONDecodeError: Expecting ',' delimiter: line 1 column 8 (char 7)

    Using ``load_json`` together with ``itemize`` allows the processing of
    JSON-lines data. ``itemize`` will yield a single item for each line and
    ``load_json`` will convert it into a JSON-object.

    Note: JSON-decoding is slightly faster if the items of type ``str``. Items
    of type ``bytes`` or ``bytearray`` will work as well, but processing might
    be slower.

    Parameters
    ----------
    iterable: Iterable[bytes | str]
        The iterable that yields the JSON-strings or -bytestrings that should be
        parsed and converted into JSON-objects

    Yields
    ------
    Any
        The JSON-object that are generated from the data yielded by ``iterable``

    Raises
    ------
    json.decoder.JSONDecodeError
        If the data yielded by ``iterable`` is not a valid JSON-string
    """
    for json_string in iterable:
        yield json.loads(json_string)


def load_json_with_flag(
    iterable: Iterable[bytes | str],
) -> Generator[tuple[Any | json.decoder.JSONDecodeError, bool], None, None]:
    """Convert items from ``iterable`` into JSON objects and a success flag

    ``load_json_with_flag`` works analogous to ``load_json``, but reports
    success and failure differently.

    On successful conversion to a JSON-object, ``load_json_with_flag`` will
    yield a tuple of two elements. The first element contains the JSON-object,
    the second element is ``True``.

    If the conversion to a JSON-object fails, ``load_json_with_flag`` will
    yield a tuple of two elements, where the first element contains the
    ``json.decoder.JSONDecodeError`` that was raised during conversion, and the
    second element is ``False``:

    .. code-block:: python

        >>> from datasalad.itertools import load_json, load_json_with_flag
        >>> tuple(load_json_with_flag(['{"b": 2}']))
        (({'b': 2}, True),)
        >>> tuple(load_json_with_flag(['{"d": 4']))   # Faulty JSON-encoding
        ((JSONDecodeError("Expecting ',' delimiter: line 1 column 8 (char 7)"), False),)

    Parameters
    ----------
    iterable: Iterable[bytes | str]
        The iterable that yields the JSON-strings or -bytestrings that should be
        parsed and converted into JSON-objects

    Yields
    ------
    tuple[Any | json.decoder.JSONDecodeError, bool]
        A tuple containing of a decoded JSON-object and ``True``, if the JSON
        string could be decoded correctly. If the JSON string  could not be
        decoded correctly, the tuple will contain the
        ``json.decoder.JSONDecodeError`` that was raised during JSON-decoding
        and ``False``.
    """
    for json_string in iterable:
        try:
            yield json.loads(json_string), True
        except json.decoder.JSONDecodeError as e:
            yield e, False



from __future__ import annotations

import sys

import pytest

from ..exception import CommandError


def test_CommandError_minial():
    with pytest.raises(TypeError):
        # we need a command
        CommandError()


def test_CommandError_str_repr() -> None:
    unicode_out = 'ΔЙקم๗あösdohabcdefdf0j23d2däüß§!אؠॵΔЙקم๗あöäüß§!אؠॵℱ⏰︻𐂃𐃍'
    cp1252_out = b'\xe4\xd6\xc4\xdf\xa7'
    testcases = [
        (
            CommandError('mycmd'),
            "Command 'mycmd' errored with unknown exit status",
            "CommandError('mycmd')",
        ),
        (
            CommandError(['mycmd', 'arg0']),
            "Command ['mycmd', 'arg0'] errored with unknown exit status",
            "CommandError(['mycmd', 'arg0'])",
        ),
        (
            CommandError(
                cmd='<cmd>',
                msg='<msg>',
                returncode=1,
                stdout='<stdout>',
                stderr='<stderr>',
                cwd='<cwd>',
            ),
            "Command '<cmd>' returned non-zero exit status 1 at CWD <cwd> [<msg>] [stderr: <stderr>]",
            "CommandError('<cmd>', msg='<msg>', returncode=1, stdout='<stdout>', "
            "stderr='<stderr>', cwd='<cwd>')",
        ),
        (
            CommandError('mycmd', stdout=unicode_out),
            "Command 'mycmd' errored with unknown exit status",
            f"CommandError('mycmd', stdout='{unicode_out[:20]}<... +14 chars>{unicode_out[-20:]}')",
        ),
        (
            CommandError('mycmd', stdout=cp1252_out),
            "Command 'mycmd' errored with unknown exit status",
            "CommandError('mycmd', stdout=b'<5 bytes>')",
        ),
        (
            CommandError('mycmd', returncode=6),
            "Command 'mycmd' returned non-zero exit status 6",
            "CommandError('mycmd', returncode=6)",
        ),
        (
            CommandError('mycmd', returncode=-234),
            "Command 'mycmd' died with unknown signal 234",
            "CommandError('mycmd', returncode=-234)",
        ),
        (
            CommandError('mycmd', stderr=unicode_out),
            f"Command 'mycmd' errored with unknown exit status [stderr: {unicode_out}]",
            f"CommandError('mycmd', stderr='{unicode_out[:20]}<... +14 chars>{unicode_out[-20:]}')",
        ),
    ]
    if not sys.platform.startswith('win'):
        testcases.extend(
            (
                # decode signal name, like `subprocess.CalledProcessError` does
                (
                    CommandError('mycmd', returncode=-6),
                    "Command 'mycmd' died with SIGABRT",
                    "CommandError('mycmd', returncode=-6)",
                ),
                # this is done on non-windows, because it uses a CP1252
                # encoded string to provoke a decoding error
                (
                    CommandError('mycmd', stderr=cp1252_out),
                    "Command 'mycmd' errored with unknown exit status "
                    '[stderr: <undecodable 5 bytes>]',
                    "CommandError('mycmd', stderr=b'<5 bytes>')",
                ),
            )
        )
    for ce, _str, _repr in testcases:
        assert str(ce) == _str
        assert repr(ce) == _repr


def check_reraise_CommandError_with_msg():
    try:
        # some call raises an original exception
        raise CommandError('mycmd')  # noqa: TRY301, EM101
    except CommandError as e:
        # a wrapping try/except can be used to add context info
        # or a hint on a probably cause to the exception
        e.msg = 'context info or hint'
        raise


def test_CommandError_context_msg():
    with pytest.raises(CommandError) as cmderr:
        check_reraise_CommandError_with_msg()
    assert cmderr.value.cmd == 'mycmd'
    assert cmderr.value.msg == 'context info or hint'
    assert (
        str(cmderr.value) == "Command 'mycmd' errored with unknown exit status "
        '[context info or hint]'
    )



from __future__ import annotations

from collections import deque
from collections.abc import Generator
from contextlib import contextmanager
from subprocess import PIPE, Popen
from threading import Thread
from typing import (
    TYPE_CHECKING,
    Iterable,
)

if TYPE_CHECKING:
    from os import PathLike

from datasalad.runners import CommandError

# Errno22 indicates an IO failure with a
# file descriptor
ERRCODE_IO_FAILURE = 22


class OutputFrom(Generator):
    def __init__(self, stdout, stderr_deque, chunk_size=65536):
        self.stdout = stdout
        self.stderr_deque = stderr_deque
        self.chunk_size = chunk_size
        self.returncode = None

    def send(self, _):
        chunk = self.stdout.read(self.chunk_size)
        if not chunk:
            raise StopIteration
        return chunk

    def throw(self, typ, value=None, traceback=None):
        return super().throw(typ, value, traceback)


@contextmanager
def iterable_subprocess(
    program: list[str],
    input_chunks: Iterable[bytes],
    chunk_size: int = 65536,
    cwd: PathLike | str | None = None,
    bufsize: int = -1,
):
    """Subprocess execution context manager with iterable IO

    This context starts a thread that populates the subprocess's standard
    input. It also starts a threads that reads the process's standard error.
    Otherwise we risk a deadlock - there is no output because the process is
    waiting for more input.

    This itself introduces its own complications and risks, but hopefully
    mitigated by having a well defined start and stop mechanism that also avoid
    sending data to the process if it's not running

    To start, i.e. on entry to the context from client code

    - The process is started
    - The thread to read from standard error is started
    - The thread to populate input is started

    When running:

    - The standard input thread iterates over the input, passing chunks to the
      process
    - While the standard error thread fetches the error output
    - And while this thread iterates over the processe's output from client
      code in the context

    To stop, i.e. on exit of the context from client code

    - This thread closes the process's standard output
    - Wait for the standard input thread to exit
    - Wait for the standard error thread to exit
    - Wait for the process to exit

    By using context managers internally, this also gives quite strong
    guarantees that the above order is enforced to make sure the thread doesn't
    send data to the process whose standard input is closed and so we don't get
    ``BrokenPipe`` errors

    Writing to the process can result in a ``BrokenPipeError``. If this then
    results in a non-zero code from the process, the process's standard error
    probably has useful information on the cause of this. However, the non-zero
    error code happens after ``BrokenPipeError``, so propagating "what happens
    first" isn't helpful in this case.  So, we re-raise ``BrokenPipeError`` as
    ``_BrokenPipeError`` so we can catch it after the process ends to then
    allow us to branch on its error code:

    - if it's non-zero raise a ``CommandError`` containing its standard error
    - if it's zero, re-raise the original ``BrokenPipeError``

    >>> # regular execution, no input iterable
    >>> with iterable_subprocess(['printf', 'test'], []) as proc:
    ...     for chunk in proc:
    ...         print(chunk)
    b'test'
    >>> # feed subprocess stdin from an iterable
    >>> with iterable_subprocess(['cat'], [b'test']) as proc:
    ...     for chunk in proc:
    ...         print(chunk)
    b'test'
    """

    class _BrokenPipeError(Exception):
        pass

    @contextmanager
    def thread(target, *args):
        exception = None

        def wrapper():
            nonlocal exception
            try:
                target(*args)
            except BaseException as e:  # noqa: BLE001
                # ok to catch any exception here, we are reporting them
                exception = e

        t = Thread(target=wrapper)

        def start():
            t.start()

        def join():
            if t.ident:
                t.join()
            return exception

        yield start, join

    def input_to(stdin):
        try:
            for chunk in input_chunks:
                try:
                    stdin.write(chunk)
                except BrokenPipeError:
                    raise _BrokenPipeError  # noqa B904
                except OSError as e:
                    if e.errno != ERRCODE_IO_FAILURE:
                        # maybe process is dead already
                        raise _BrokenPipeError  # noqa B904
                    # no idea what this could be, let it bubble up
                    raise
        finally:
            try:
                stdin.close()
            except BrokenPipeError:
                raise _BrokenPipeError  # noqa B904
            except OSError as e:
                # silently ignore Errno22, which happens on
                # windows when trying to interacted with file descriptors
                # associated with a process that exited already
                if e.errno != ERRCODE_IO_FAILURE:
                    raise

    def keep_only_most_recent(stderr, stderr_deque):
        total_length = 0
        while True:
            chunk = stderr.read(chunk_size)
            total_length += len(chunk)
            if not chunk:
                break
            stderr_deque.append(chunk)
            if total_length - len(stderr_deque[0]) >= chunk_size:
                total_length -= len(stderr_deque[0])
                stderr_deque.popleft()

    def raise_if_not_none(exception):
        if exception is not None:
            raise exception from None

    proc: Popen | None = None
    stderr_deque: deque[bytes] = deque()
    chunk_generator = None
    exception_stdin = None
    exception_stderr = None

    try:
        with Popen(  # nosec - all arguments are controlled by the caller
            program,
            stdin=PIPE,
            stdout=PIPE,
            stderr=PIPE,
            cwd=cwd,
            bufsize=bufsize,
        ) as proc, thread(
            keep_only_most_recent,
            proc.stderr,
            stderr_deque,
        ) as (start_t_stderr, join_t_stderr), thread(
            input_to,
            proc.stdin,
        ) as (start_t_stdin, join_t_stdin):
            try:
                start_t_stderr()
                start_t_stdin()
                chunk_generator = OutputFrom(proc.stdout, stderr_deque, chunk_size)
                yield chunk_generator
            except BaseException:
                proc.terminate()
                raise
            finally:
                if TYPE_CHECKING:
                    assert proc is not None
                    assert proc.stdout is not None
                proc.stdout.close()
                exception_stdin = join_t_stdin()
                exception_stderr = join_t_stderr()

            raise_if_not_none(exception_stdin)
            raise_if_not_none(exception_stderr)

    except _BrokenPipeError as e:
        if TYPE_CHECKING:
            assert proc is not None
        if chunk_generator:
            chunk_generator.returncode = proc.returncode
        if proc.returncode == 0:
            if TYPE_CHECKING:
                assert isinstance(e.__context__, BrokenPipeError)  # noqa PT017
            raise e.__context__ from None
    except BaseException:
        if TYPE_CHECKING:
            assert proc is not None
        if chunk_generator:
            chunk_generator.returncode = proc.returncode
        raise

    if TYPE_CHECKING:
        assert chunk_generator is not None
    chunk_generator.returncode = proc.returncode
    if proc.returncode:
        raise CommandError(
            cmd=program,
            returncode=proc.returncode,
            stderr=b''.join(stderr_deque)[-chunk_size:],
            cwd=cwd,
        )



"""Get complete items from input chunks"""

from __future__ import annotations

from typing import (
    Generator,
    Iterable,
    TypeVar,
)

__all__ = ['itemize']


T = TypeVar('T', str, bytes, bytearray)


def itemize(
    iterable: Iterable[T],
    sep: T | None,
    *,
    keep_ends: bool = False,
) -> Generator[T, None, None]:
    """Yields complete items (only), assembled from an iterable

    This function consumes chunks from an iterable and yields items defined by
    a separator. An item might span multiple input chunks.  Input (chunks) can
    be ``bytes``, ``bytearray``, or ``str`` objects.  The result type is
    determined by the type of the first input chunk. During its runtime, the
    type of the elements in ``iterable`` must not change.

    Items are defined by a separator given via ``sep``. If ``sep`` is ``None``,
    the line-separators built into ``str.splitlines()`` are used, and each
    yielded item will be a line. If ``sep`` is not `None`, its type must be
    compatible to the type of the elements in ``iterable``.

    A separator could, for example, be ``b'\\n'``, in which case the items
    would be terminated by Unix line-endings, i.e. each yielded item is a
    single line. The separator could also be, ``b'\\x00'`` (or ``'\\x00'``),
    to split zero-byte delimited content, like the output of
    ``git ls-files -z``.

    Separators can be longer than one byte or character, e.g. ``b'\\r\\n'``, or
    ``b'\\n-------------------\\n'``.

    Content after the last separator, possibly merged across input chunks, is
    always yielded as the last item, even if it is not terminated by the
    separator.

    Performance notes:

    - Using ``None`` as a separator  (splitlines-mode) is slower than providing
      a specific separator.
    - If another separator than ``None`` is used, the runtime with ``keep_end=False`` is faster than with ``keep_end=True``.

    Parameters
    ----------
    iterable: Iterable[str | bytes | bytearray]
        The iterable that yields the input data
    sep: str | bytes | bytearray | None
        The separator that defines items. If ``None``, the items are
        determined by the line-separators that are built into
        ``str.splitlines()``.
    keep_ends: bool
        If `True`, the item-separator will remain at the end of a
        yielded item. If `False`, items will not contain the
        separator. Preserving separators implies a runtime cost, unless the separator is ``None``.

    Yields
    ------
    str | bytes | bytearray
        The items determined from the input iterable. The type of the yielded
        items depends on the type of the first element in ``iterable``.

    Examples
    --------

    .. code-block:: python

        >>> from datasalad.itertools import itemize
        >>> with open('/etc/passwd', 'rt') as f:                            # doctest: +SKIP
        ...     print(tuple(itemize(iter(f.read, ''), sep=None))[0:2])      # doctest: +SKIP
        ('root:x:0:0:root:/root:/bin/bash',
         'systemd-timesync:x:497:497:systemd Time Synchronization:/:/usr/sbin/nologin')
        >>> with open('/etc/passwd', 'rt') as f:                            # doctest: +SKIP
        ...     print(tuple(itemize(iter(f.read, ''), sep=':'))[0:10])      # doctest: +SKIP
        ('root', 'x', '0', '0', 'root', '/root',
         '/bin/bash\\nsystemd-timesync', 'x', '497', '497')
        >>> with open('/etc/passwd', 'rt') as f:                                        # doctest: +SKIP
        ...     print(tuple(itemize(iter(f.read, ''), sep=':', keep_ends=True))[0:10])  # doctest: +SKIP
        ('root:', 'x:', '0:', '0:', 'root:', '/root:',
         '/bin/bash\\nsystemd-timesync:', 'x:', '497:', '497:')
    """
    if sep is None:
        yield from _split_lines(iterable, keep_ends=keep_ends)
    else:
        yield from _split_items_with_separator(
            iterable,
            sep=sep,
            keep_ends=keep_ends,
        )


def _split_items_with_separator(
    iterable: Iterable[T],
    sep: T,
    *,
    keep_ends: bool = False,
) -> Generator[T, None, None]:
    assembled = None
    for chunk in iterable:
        if not assembled:
            assembled = chunk
        else:
            assembled += chunk
        items = assembled.split(sep=sep)
        if len(items) == 1:
            continue

        assembled = None if assembled.endswith(sep) else items[-1]
        items.pop(-1)
        if keep_ends:
            for item in items:
                yield item + sep
        else:
            yield from items

    if assembled:
        yield assembled


def _split_lines(
    iterable: Iterable[T],
    *,
    keep_ends: bool = False,
) -> Generator[T, None, None]:
    assembled = None
    for chunk in iterable:
        if not assembled:
            assembled = chunk
        else:
            assembled += chunk
        # We don't know all elements on which python splits lines, therefore we
        # split once with ends and once without ends. Lines that differ have no
        # ending
        lines_with_end = assembled.splitlines(keepends=True)
        lines_without_end = assembled.splitlines(keepends=False)
        if lines_with_end[-1] == lines_without_end[-1]:
            assembled = lines_with_end[-1]
            lines_with_end.pop(-1)
            lines_without_end.pop(-1)
        else:
            assembled = None
        if keep_ends:
            yield from lines_with_end
        else:
            yield from lines_without_end

    if assembled:
        yield assembled
